######################################################################
#:CQCC::IOPtkManip
#
# Licensed Materials - Property of IBM
# Rational ClearCase
# (C) Copyright IBM Corp. 2001, 2008. All Rights Reserved.
# US Government Users Restricted Rights
# Use, duplication or disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
#

=head1 NAME 

CQCC::IOPtkManip - IOPtk widget class for managing data interaction

=head1 CLASS DESCRIPTION

This class provides wrapper objects for a number of commonly used
data manipulator widgets, hiding the implementation details and
handling message translation and Parm object interaction. 

As a subclass of IOPtkWidget, it inherits some basic attribute 
management and initialization functionality. In addition this class
manages interaction with Parm objects or other values. 

How the IOPtkManip object looks and behaves depends on its "subtype"
which provides a lightweight subclassing feature. Each subtype has an 
initialization function that uses Perl/TK calls to define the widget;
the subtype may have additional supporting methods to implement its
behavior. 

The currently supported subtypes include IOButton, IOLabel, IOListbox,
IOListboxGrid, IOListboxTree, and IOScrolledLabel. See the method
documentation on each for more information.

Data in the manipulator widget is accessed using the ManageValues
method which handles getting or setting values or ranges. This 
should be called through the CQCC::IOPerlTK::ManipManageValues
interface method. 

Manipulators are created by the new method but this should be 
called through the CQCC::IOPerlTK::Manipulator interface method.

The IOPtkPanel ClassTest method provides an extensive test of all 
of the IOPtkWidget classes. 

SUPPORT POLICY: See TriggerCQCC.pm for the Rational Support Policy. 

=head1 METHODS

=cut

######################################################################
#::Package declarations
######################################################################

require 5.006; 

package CQCC::IOPtkManip;
use CQCC::IOPtkWidget;
@ISA = qw( CQCC::IOPtkWidget );

use Tk 800.022;
use Tk::DialogBox;
require Tk::HList;
require Tk::ItemStyle;
require Tk::Label;
require Tk::LabFrame;
require Tk::LabEntry;
require Tk::Pane;
require Tk::Tree;

use strict;

use CQCC::IOBase;
use CQCC::Parm;
use CQCC::MsgCat;
use CQCC::IOPtkResizeButton;

sub Version { return '2.1' };

######################################################################
#::Class Globals 
######################################################################

my %listboxTreeLookup = (); # Used to determine widget from filename

######################################################################
#::function ConvertValues
# Handles conversion and local error reporting

sub ConvertValues {
    my ($routineName, $dst_codepage, @values) = @_;
    my $src_codepage = ($dst_codepage eq "CORE_CODEPAGE") ? 
                       "IO_CODEPAGE" : "CORE_CODEPAGE";
    my ($status,@convertedValues,$v);
    for $v (@values) {
        my ($status,@vList) = &IOConvertString($v,$src_codepage, $dst_codepage);
        if ($status) {
            IOMessage("FATAL", ["IOPTK_E_CONVERSION_ERROR",$routineName,$status]);
        }
        push @convertedValues, (ref($v) ? \@vList : @vList);
    }

    return @convertedValues;
}

######################################################################
#::method IOButton

=over 6

=item C<IOButton subtype>

Calling sequence using an instance of IOPerlTK called $IO:

    $IO->Manipulator(name, parent, "IOButton", key-values)>

This subtype provides a generic button manipulator widget. It uses the
Perl/TK Button object directly after translating its TITLE and setting
a callback.

The callback is similar to the IOPtkMenu callbacks. The 
CALLBACK_OBJECT attribute defines an object and the CALLBACK
attribute defines the method name or a list reference containing
the method name and argument list. 

This is a sample of a IOButton being created: 

    $IO->Manipulator("Database", "Frame2", "IOButton",
               TITLE=>["GUI_Database"],
               TK=>[ -background => 'white' ],
               CALLBACK_OBJECT => $self,
               CALLBACK => "Dialog_SelectDatabase");

=back

=cut

######################################################################

sub IOButton {
    my ($self) = @_;

    ######################################### Perl/TK code
    $self->{WIDGET} = $self->{PARENT_WIDGET}->
        Button( -text => $self->{TITLEMSG},
                -command => [ \&CQCC::IOPtkWidget::DoCallback, $self,
                              $self->{TITLE}, $self->{CALLBACK} ],
                $self->TKoptions()
               )->pack($self->PackArgs());
    ######################################## 

    return $self;
}

######################################################################
#::method IOLabel

=over 6


=item C<IOLabel subtype>

Calling sequence using an instance of IOPerlTK called $IO:

    $IO->Manipulator(name, parent, "IOLabel", key-values)>

This subtype method defines a generic label widget that shows the
current value of a Parm object or a raw value passed in through the
VALUE attribute.

This is a sample of making an IOLabel called DatabaseName based on the
value of a Parm called CQCC_DATABASE.

    $IO->Manipulator("DatabaseName", "Frame2", "IOLabel",
                     PARM=>"CQCC_DATABASE",
                     TK => [ -background => 'white', 
                             -relief => 'sunken' ],

The TK attribute passes through one or more Perl/TK attribute
pairs to customize the widget. 

The VALUE attribute may be used in place of the PARM attribute
to provide a direct data value. 

=back

=cut

######################################################################

sub IOLabel {
    my ($self) = @_;

    ######################################### Perl/TK code
    $self->{WIDGET} = $self->{PARENT_WIDGET}->
        Label(-textvariable => $self->{VAR}, 
              $self->TKoptions()
             )->pack($self->PackArgs(-side=>"left", -anchor=>"n"));
    ########################################

    return $self;
}

######################################################################
#::method IOListbox

=over 6

=item C<IOListbox subtype>

Calling sequence using an instance of IOPerlTK called $IO:

    $IO->Manipulator(name, parent, "IOListbox", key-values)>

This subtype defines a listbox manipulator widget which provides a
scrolling list of options to select from.

The widget is built from an outer frame and includes a title label if
the TITLE attribute is set. The main widget is a Perl/TK Scrolled
Listbox.

The RANGE attribute controls the possible selections and the VALUE
is the list of selections actually made by the user. 

The MODE attribute controls the selection mode ("single", "browse",
"multiple"). 

If HEIGHT is provided it controls the height of the IOListbox in
lines.

This is a sample of a declaration of an IOListbox using the
Manipulator method of IOPerl/TK.

    $IO->Manipulator("List1", "Page1", "IOListbox",
                     TITLE=>"List", 
                     MODE => "single",
                     HEIGHT => 10,
                     RANGE => ["a","b","c"],
                     PACK=>[-side => "left", -fill => 'both']);

The IOListbox manipulator uses custom internal functions to modify or
retrieve its range or value settings. 

=back

=cut

######################################################################

sub IOListbox { 
    my ($self) = @_;
    if (! defined $self->{MODE}) {
       $self->{MODE} = "single";
    } 

    ######################################### Perl/TK code
    $self->{FRAME} = $self->{PARENT_WIDGET}->
        Frame($self->TKoptions()
             )->pack($self->PackArgs( -fill => "both", -expand=>1 ));

    if ($self->{TITLEMSG}) { 
        $self->{LABEL} = $self->{FRAME}->
            Label(-text => $self->{TITLEMSG}
                 )->pack(-side =>"top", -anchor => "w");
    } 
    my @heightList; 
    if ($self->{HEIGHT}) { 
        @heightList = (-height => $self->{HEIGHT});
    }
    $self->{WIDGET} = $self->{FRAME}->
        Scrolled("Listbox", 
                 -background => "white",
                 -scrollbars =>"osoe",
                 -selectmode => $self->{MODE},
                 @heightList
                )->pack(-side => "top", -expand => 1, -fill=>"both");
    #########################################

    $self->{GETVALUE} = "_IOListboxGetValue";
    $self->{UPDATE} = "_IOListboxRefresh";
    $self->_IOListboxRefresh();
    
    return $self;
} 

######################################################################
#::method _IOListboxGetValue
#
# This is a custom internal function used by the IOListbox manipulator
# to get its current value list.
######################################################################

sub _IOListboxGetValue { 
    my ($self) = @_;
    my @answers;

    ######################################### Perl/TK code
    my @listSelection = $self->{WIDGET}->curselection;
    my @listOptions = $self->{WIDGET}->get(0,'end');
    $self->{WIDGET}->selectionClear(0,'end');
    ########################################

    for (@listSelection) { 
        push @answers, $listOptions[$_];
    }

    @answers = &ConvertValues("IOPtkManip::_IOListboxGetValue", 
                              "CORE_CODEPAGE", @answers);
    $self->{VALUE} = \@answers;
    return;
}

######################################################################
#::method _IOListboxRefresh
#
# This is a custom internal function used by the IOListbox manipulator
# to update its range of allowed selections.
######################################################################

sub _IOListboxRefresh {
    my ($self) = @_;
    my @choices = $self->{RANGE} ? @{ $self->{RANGE} } : ();
    my $widget = $self->{WIDGET};

    @choices = &ConvertValues("IOPtkManip::_IOListboxRefresh", 
                              "IO_CODEPAGE", @choices);

    ######################################### Perl/TK code
    $widget->delete(0, 'end');
    $widget->insert('end', @choices);
    ########################################
}

######################################################################
#::method IOListboxGrid

=over 6

=item C<IOListboxGrid subtype>

Calling sequence using an instance of IOPerlTK called $IO:

    $IO->Manipulator(name, parent, "IOListboxGrid", key-values)>

This subtype method defines a listbox widget which provides a
scrolling list of column based data to select from.

The widget is built wrapped inside a frame and includes a title label
if the TITLE attribute is set. The main widget is a Perl/TK Scrolled
Listbox with special column header buttons used to provide resize
control and sorting for the columns.

The RANGE attribute controls the possible selections and the VALUE
is the list of selections actually made by the user.

The MODE attribute controls the selection mode ("single", "browse",
"multiple").

If HEIGHT is provided it controls the height of the IOListboxGrid in
lines.

The first row of data is assumed to be the column heading titles. 

This is a sample of a declaration of an IOListboxGrid using the
Manipulator method of IOPerl/TK.

    my @data = ( [ "ID", "Heading", "State" ],
                 [ "SAMPL0056", "Test heading 4", "Submit" ]);

    $IO->Manipulator("MyListboxGrid", "Page2", "IOListboxGrid",
                     TITLE=>"IOListboxGrid Demo", 
                     MODE => "single",
                     HEIGHT => 10,
                     RANGE => \@data,
                     PACK=>[-side => "left", -fill => 'both']);

The widget will only display as many columns as the data contains but
must be prebuilt with the maximum number of columns expected so that
the widget can be reused with new data. The MAX_COLUMNS attribute
determines this limit and defaults to 20.

The IOListboxGrid manipulator uses custom internal functions to modify
or retrieve its range or value settings.

=back

=cut

######################################################################

sub IOListboxGrid { 
    my ($self) = @_;
    if (! defined $self->{MODE}) {
       $self->{MODE} = "single";
    } 

    if (! defined $self->{MAX_COLUMNS}) {
       $self->{MAX_COLUMNS} = 20;
    } 

    if (! defined $self->{COLUMN_NAMES}) {
        my @colNames;
        for (0 .. ($self->{MAX_COLUMNS} - 1)) { 
            $colNames[$_] = "";
        }
        @colNames = &ConvertValues("IOPtkManip::IOListboxGrid", 
                                   "IO_CODEPAGE", @colNames);
        $self->{COLUMN_NAMES} = \@colNames;
    }
    $self->{OLD_COLUMN_NAMES} = ["###UNSET###"];

    ######################################### Perl/TK code

    $self->{FRAME} = $self->{PARENT_WIDGET}->
        Frame($self->TKoptions()
             )->pack($self->PackArgs( -fill => "both", -expand=>1 ));

    if ($self->{TITLEMSG}) { 
        $self->{LABEL} = $self->{FRAME}->
            Label(-text => $self->{TITLEMSG}
                 )->pack(-side =>"top", -anchor => "w");
    } 
    my @heightList; 
    if ($self->{HEIGHT}) { 
        @heightList = (-height => $self->{HEIGHT});
    }
    my $listBox = $self->{WIDGET} = $self->{FRAME}->
        Scrolled("HList", 
                 -background => "white",
                 -columns => $self->{MAX_COLUMNS},
                 -header => 1,
                 -itemtype => 'text',
                 -scrollbars =>"osoe",
                 -selectmode => $self->{MODE},
                 @heightList,
                )->pack(-side => "top", -expand => 1, 
                        -fill => "both");

    my $headerstyle =
        $listBox->ItemStyle('window', -padx => 0, -pady => 0);

    my $scrolledHList = $listBox->Subwidget('scrolled');
    
    my @columnButtons;

    for (0 .. ($self->{MAX_COLUMNS} - 1)) { 
        my $columnButton = 
            $listBox->IOPtkResizeButton(-column => $_, 
                                        -relief => 'flat', 
                                        -widget => \$scrolledHList,
                                        -justify => 'left',
                                        -command => 
                                           [ \&_IOListboxGridSort, 
                                             $self, $_ ]
                                        );
        $listBox->header('create', $_,
                         -itemtype => 'window',
                         -widget => $columnButton,
                         -style => $headerstyle);
        $self->{WIDGET}->columnWidth($_, "");
        push @columnButtons, $columnButton;
    }
    $self->{COLUMN_BUTTONS} = \@columnButtons;
    
    #########################################

    $self->{GETVALUE} = "_IOListboxGridGetValue";
    $self->{UPDATE} = "_IOListboxGridRefresh";
    $self->_IOListboxGridRefresh();
    
    return $self;
} 

######################################################################
#::method _IOListboxGridGetValue
#
# This is a custom internal function used by the IOListboxGrid widget
# to get its current selections. 
######################################################################

sub _IOListboxGridGetValue { 
    my ($self) = @_;
    my @choices = $self->{RANGE} ? @{ $self->{RANGE} } : ();
    my @answers;

    ######################################### Perl/TK code
    my @listSelection = $self->{WIDGET}->selectionGet();
    $self->{WIDGET}->anchorClear();
    $self->{WIDGET}->selectionClear();
    ########################################

    # Add one to skip over the 
    for (@listSelection) { 
        push @answers, $choices[$_];
    }

    $self->{VALUE} = \@answers;
    return;
}

######################################################################
#::method _IOListboxGridRefresh
#
# This is a custom internal function used by the IOListboxGrid widget
# to update its range of allowed selections.
######################################################################

sub _IOListboxGridRefresh {
    my ($self) = @_;
    my @choices = $self->{RANGE} ? @{ $self->{RANGE} } : ();

    @choices = &ConvertValues("IOPtkManip::_IOListboxGridRefresh",
                              "IO_CODEPAGE",@choices);

    my $widget = $self->{WIDGET};
    my @colNames = @{ $self->{COLUMN_NAMES} };

    # Get latest column names from query and retain them in case
    # we later clear out all the entries

    if (@choices) { 
        @colNames = @{ $choices[0] };
        @colNames = &ConvertValues("IOPtkManip::_IOListboxGridRefresh",
                                   "IO_CODEPAGE", @colNames);
        $self->{COLUMN_NAMES} = \@colNames;
    }

    # Determine if the column names changed - don't resize
    # columns if they are the same to avoid overriding user
    # width adjustments. 

    my @oldColNames = @{ $self->{OLD_COLUMN_NAMES} };
    my $newNames = 1;

    if ($#colNames == $#oldColNames) {
        my $sameNames = 1;
        for (0 .. $#colNames) { 
            if ($colNames[$_] ne $oldColNames[$_]) {
                $sameNames = 0;
                last;
            }
        }
        if ($sameNames) {
            $newNames = 0;
        }
    }

    ######################################### Perl/TK code
    $widget->delete("all");

    my @columnButtons = @{ $self->{COLUMN_BUTTONS} };

    if ($newNames) { 
        for (0 .. ($self->{MAX_COLUMNS} - 1)) { 
            $columnButtons[$_]->configure(-text => $colNames[$_]);
            $self->{WIDGET}->columnWidth($_,
                length($colNames[$_]) > 0 ? "" : 0);
        }
        $self->{OLD_COLUMN_NAMES} = \@colNames;
    }
    
    for (1 .. $#choices) { 
        my @fields = @{ $choices[$_] };
        my $line = $_;
        $widget->add($line);
        my $numColumns = ($#fields > ($self->{MAX_COLUMNS}-1)
                          ? ($self->{MAX_COLUMNS}-1) : $#fields);
        for (0 .. $numColumns) {
            $widget->itemCreate($line, $_, -text => $fields[$_]);
        }
    }
    $widget->pack();
    ########################################
}

######################################################################
#::method _IOListboxGridSort
#
# This is a custom internal function used by the IOListboxGrid widget
# to handle the sort operation invoked by clicking the column heading.
######################################################################

sub _IOListboxGridSort {
    my ($self, $column) = @_;
    my @choices = $self->{RANGE} ? @{ $self->{RANGE} } : ();
    my $widget = $self->{WIDGET};
    
    # Build hashkey list from nth column of data with numbers
    # Produce a new data list from sorted data with names
    # at the top. Remember that the first row is column names.
    
    my %sortKeys;
    my $keySep = "<#>";

    for (1 .. $#choices) {
        my @fields = @{ $choices[$_] };
        # Make sure key is unique, favor original order
#L10N: Probably most user visible sorting operation
        my $sortKey = $fields[$column] . $keySep . $fields[0];
        $sortKeys{ $sortKey } = $_;               
    }
    
    my @sortedChoices;
    push @sortedChoices, $choices[0];
    my @sorted = sort keys %sortKeys;

    # Ascending sort the first time, descending the second time

    my $saveSortKey;
    if (defined $self->{SORTKEY} && $self->{SORTKEY} == $column) {
        @sorted = reverse @sorted;
        $self->{SORTKEY} = undef;
    }
    else {
        $saveSortKey = $column
    } 
    
    for (@sorted) {
        my ($key,$order) = split($keySep, $sortKeys{ $_ });
        push @sortedChoices, $choices[ $key ];
    }
    $self->ManageValues("SETRANGE",\@sortedChoices);
    $self->{SORTKEY} = $saveSortKey;  # save to recognize reverse sort
}

######################################################################
#::method IOListboxTree

=over 6

=item C<IOListboxTree subtype>

Calling sequence using an instance of IOPerlTK called $IO:

    $IO->Manipulator(name, parent, "IOListboxTree", key-values)>

This subtype method defines a listbox widget which provides a
scrolling list showing a hierarchical tree.

The widget is built wrapped inside a frame and includes a title label
if the TITLE attribute is set. The main widget is a Perl/TK Scrolled
Tree widget. 

The RANGE attribute controls the possible selections. The value of
the widget is the list of selections actually made by the user.

The MODE attribute controls the selection mode ("single", "browse",
"multiple").

If HEIGHT is provided it controls the height of the IOListboxTree in
lines.

If DIALOG_BOX and DIALOG_BUTTON are provided, it will attempt to 
invoke the DIALOG_BUTTON of DIALOG when a double click happens on a
leaf node. 

This is a sample of a declaration of a IOListboxTree using the
Manipulator method of IOPerl/TK. The tree structure is determined from
the pathnames of the data items. The path elements are delimited by 
the SEP attribute, which defaults to "/".

Intermediate branches are added as needed, based on the path. In the
example below, the Work and Work/Personal Queries branches would be
created automatically for the first item in the list. Only original
data elements are selectable - selecting the intermediate branch is
automatically unselected. 

    my @q = ("Work/Personal Queries/Report 0",
             "Work/Public Queries/Aging Charts/Closed 30 Day report",
             "Work/Public Queries/Aging Charts/Open/7 Day report");

    $IO->Manipulator("MyListboxTree", "Page2", "IOListboxTree",
                     TITLE=>"IOListboxTree Demo", 
                     MODE => "single",
                     HEIGHT => 10,
                     RANGE => \@q,
                     PACK=>[-side => "left", -fill => 'both']);

The IOListboxTree manipulator uses custom internal functions to modify
or retrieve its range or value settings.

=back

=cut


######################################################################

sub IOListboxTree { 
    my ($self) = @_;
    if (! defined $self->{MODE}) {
       $self->{MODE} = "single";
    } 
    if (! defined $self->{SEPARATOR}) {
       $self->{SEPARATOR} = "/";
    } 

    if (! defined $self->{CLOSE_LEVEL}) {
       $self->{CLOSE_LEVEL} = 1;
    } 

    ######################################### Perl/TK code
    $self->{FRAME} = $self->{PARENT_WIDGET}->
        Frame($self->TKoptions()
             )->pack($self->PackArgs( -fill => "both", -expand => 1));

    if ($self->{TITLEMSG}) { 
        $self->{LABEL} = $self->{FRAME}->
            Label(-text => $self->{TITLEMSG}
                 )->pack(-side =>"top", -anchor => "w");
    } 

    my @heightList; 
    if ($self->{HEIGHT}) { 
        @heightList = (-height => $self->{HEIGHT});
    }

    my $listBox = $self->{WIDGET} = $self->{FRAME}->
        Scrolled("Tree",
                 -itemtype   => 'text',
                 -separator  => $self->{SEPARATOR},
                 -selectmode => $self->{MODE},
                 -background => "white",
                 -scrollbars =>"osoe",
                 -browsecmd=>sub { _IOListboxTreeUnselectNodes(@_);},
                 @heightList,
                )->pack(-side => "top", -expand => 1, -fill=>"both");

    $listBox->bind("<Double-Button-1>", 
         [ \&CQCC::IOPtkManip::_IOListboxTreeDoubleClick, $listBox ]);

    #########################################

    $self->{GETVALUE} = "_IOListboxTreeGetValue";
    $self->{UPDATE} = "_IOListboxTreeRefresh";
    $self->_IOListboxTreeRefresh();
    
    return $self;
} 

######################################################################
#::method _IOListboxTreeGetValue
#
# This is a custom internal function used by the IOListboxTree widget
# to get its current selections. 
######################################################################

sub _IOListboxTreeGetValue { 
    my ($self) = @_;
    my @answers;

    ######################################### Perl/TK code
    my @listSelection = $self->{WIDGET}->selectionGet();
    $self->{WIDGET}->anchorClear();
    $self->{WIDGET}->selectionClear();
    ########################################

    $self->{VALUE} = \@listSelection;
    return;
}

######################################################################
#::method _IOListboxTreeRefresh
#
# This is a custom internal function used by the IOListboxTree widget
# to update its range of allowed selections.
######################################################################

sub _IOListboxTreeRefresh {
    my ($self) = @_;
    my @choices = $self->{RANGE} ? @{ $self->{RANGE} } : ();
    my $widget = $self->{WIDGET};
    my $sep = $self->{SEPARATOR};

    ######################################### Perl/TK code
    $widget->delete("all");

    my %folderList = ();
    my @closeList;

    for ( @choices ) {
       my $fileName = $_;
       my @elements = split($sep,$fileName);
       my $i;
       # Note: loop has to use < $#elements to work
       for ($i = 0; $i < $#elements; $i++) { 
           my $folder = "";
           for (0..$i) {
               if ($folder) {
                   $folder .= $sep . $elements[$_];
               }
               else {
                   $folder = $elements[$_];
               }
           }
           if (! $folderList{$folder}) {
               $widget->add($folder, -text=>$elements[$i],
                            -data=>"NODE,$fileName");

               # close means close-able, open means open-able

               $widget->setmode($folder, "close");
               if ($i >= $self->{CLOSE_LEVEL}) {
                   push @closeList, $folder;
               }
               $folderList{$folder} = $fileName;

               # Add folder name to hash list with related widget 
               # So the browsecmd callback can determine widget
               # Not good way to do it but no alternative
               # If conflict only most recent widget is stored
               # Will keep refs to dead widgets but won't be called
               # unless user can select the folder name so no harm.

               $listboxTreeLookup{$folder} = $widget;
           }
       }
       $widget->add($fileName, -text=>$elements[$#elements],
                    -data=>"LEAF,$self->{NAME}");

       $listboxTreeLookup{$fileName} = $widget;
    }
    for (@closeList) {
       $widget->close($_);
    }
    $widget->pack(-expand => 1, -fill => "both");
    ########################################
}

######################################################################
#::method _IOListboxTreeUnselectNodes
# 
# This is a custom internal function used by the IOListboxTree widget
# to automatically unselect intermediate branch nodes that are not
# part of the original data range. 
######################################################################

sub _IOListboxTreeUnselectNodes { 
    my ($file) = @_;

    my $widget = $listboxTreeLookup{$file};
    if ($widget) { 
        my @selections = $widget->selectionGet();
        my $data = $widget->info('data',$file);
        my ($type,$name) = split(",",$data);
        if (@selections && $type eq "NODE") { 
           $widget->selectionClear();
           $widget->anchorClear();
        }
    }
}

######################################################################
#::method _IOListboxTreeDoubleClick
#
# This is a custom internal function used by the IOListboxTree widget
# to help unselect intermediate branch nodes that are not
# part of the original data range. 
######################################################################

sub _IOListboxTreeDoubleClick {
    my ($widget) = @_;

    my @selections = $widget->selectionGet();
    if (@selections) {
        my ($file) = @selections;
        my $data = $widget->info('data',$file);
        my ($type,$name) = split(",",$data);
        if ($type eq "NODE") {
           $widget->selectionClear();
           $widget->anchorClear();
        }

        # Handle double click for dialog boxes to automate "OK"
        # Involves black Perl/Tk magic:
        #    if any problems, it should fail quietly
        else {

           # BUG: Disable this code for now, causes odd mouse
           # behavior after the dialog box disappears - autoselect.
           return;

           my $self = &WidgetLookup($name);
           if ($self &&
                  $self->{DIALOG_BOX} && $self->{DIALOG_BUTTON}) {
               my $dialogObject = &WidgetLookup($self->{DIALOG_BOX});
               my $buttonMsg = 
                   $self->{MSGCAT}->Message($self->{DIALOG_BUTTON});
               if ($dialogObject && $buttonMsg) {
                   my $dwidget = $dialogObject->TKSubWidget("TOP");
                   if ($dwidget) { 
                       my @subwidgets = $dwidget->Subwidget();
                       for (@subwidgets) {
                           if ($_->name() eq "button") {
                               my $buttonWidget = $_;
                               my @textConfig = $_->configure("text");
                               for (@textConfig) { 
                                   if ($_ eq $buttonMsg) {
                                       $buttonWidget->invoke();
                                       Tk::break();
				       return;
                                   }
                               }
                           }
                       }
                   }
               }
           }                           
       }
    }
}

######################################################################
#::method IOScrolledLabel

=over 6

=item C<IOScrolledLabel subtype>

Calling sequence using an instance of IOPerlTK called $IO:

    $IO->Manipulator(name,parent,"IOScrolledLabel",key-values)>

This subtype method defines a scrolled label manipulator widget that
shows the current value of a Parm object or a raw value passed in
through the VALUE attribute. See the information on the IOLabel
subtype for more information. 

=back

=cut

######################################################################

sub IOScrolledLabel {
    my ($self) = @_;

    ######################################### Perl/TK code
    $self->{FRAME} = $self->{PARENT_WIDGET}->
        Frame(-label => $self->{TITLEMSG},
              )->pack($self->PackArgs(-side=>"left", -anchor=>"n"));

    $self->{PANE} = $self->{FRAME}->
        Scrolled("Pane", -scrollbars =>"osoe",
                )->pack(-expand => 1,-fill => "both");

    $self->{WIDGET} = $self->{PANE}->
        Label(-textvariable => $self->{VAR}, 
              $self->TKoptions())->pack(-expand => 1, -fill=>'both');
    ########################################

    return $self;
}

######################################################################
#::method ManageValues

=over 6

=item C<ManageValues(operation, value)>

This method is called to get or set the value or range of a 
manipulator widget. The currently supported operations include the
following: 

    GETRANGE - retrieves widget's current range as a list reference
    SETRANGE - sets widget range to a list reference value
    GETVALUE - retrieves widget's current range as a list
    SETVALUE - sets widget value to list reference value

The value argument is only expected by the SET* operations. 
Note: Manipulators attached to Parm objects should automatically
update their values when the parameter is changed. 

=back

=cut

######################################################################

sub ManageValues {
    my ($self,$op,$value) = @_;

    if ($op eq "GETRANGE") {
       my $range = $self->{RANGE};
       #Receiver has to check for if (ref($range))
       if ($range) {
           return @{ $range };
       }
       return ();

    }
    elsif ($op eq "SETRANGE") {   # Expects list reference
        $self->{RANGE} = $value;
        $self->{SORTKEY} = undef;
        if ($self->{UPDATE}) {
            my $op = $self->{UPDATE};
            $self->$op();
        }
    }

    elsif ($op eq "GETVALUE") {
       if ($self->{GETVALUE}) { 
          my $op = $self->{GETVALUE};
          $self->$op();
       }
       my $value = $self->{VALUE};
       if (ref($value)) {
           return @{ $value };
       }
       return ($value);
    } 

    # expects scalar or scalar reference to list, etc. 
    elsif ($op eq "SETVALUE") { 
       $self->{VALUE} = $value;
       if ($self->{UPDATE}) { 
           my $op = $self->{UPDATE};
           $self->$op();
       }
    }
    return;
}

######################################################################
#::method new

=over 6

=item C<new CQCC::IOPtkManip()>

This method builds on the IOPtkWidget constructor and adds management
for interacting with a Parm parameter or a raw VALUE by creating a
private variable. It should not be called directly but through 
CQCC::IOPerlTK::Manipulator for consistency.

=back

=cut

######################################################################

sub new { 
   my $proto = shift;
   my $class = ref($proto) || $proto;  # Object or class name

   my $self = new CQCC::IOPtkWidget(@_);

   $self->{CLASSNAME} = "CQCC::IOPtkManip";
   bless($self, $class);

   # $self->{RANGE} - Range list reference
   # $self->{VALUE} - Value or value list reference
   $self->{UPDATE} = undef;           # Update method name
   $self->{GETVALUE} = undef;         # GetValue method name

   if ($self->{PARM}) {
      $self->{VAR} = CQCC::Parm::GetParmHandle($self->{PARM}, "IO_CODEPAGE");
   }
   else {
      my $value = $self->{VALUE} ? 
                      $self->{MSGCAT}->Message($self->{VALUE}) : "";
      $self->{VAR} = \$value;
   }
   $self->NewSubType();

   return $self;
}

######################################################################
#::Tail
######################################################################

1;
